;;;; mother modular generic compiler project
; (c)2002/GPL Oskar Schirmer <oskar@scara.com>
; see file COPYING for GPL license details

; parse.scm: parser
; usage:
;   parse extended-affices-grammar-description-file <input-file >output-file
;
; the extended-affices-grammar-description-file contains a list of
; rule definitions.
;
; grammar: (rules)
; rule: ("nonterm-name" ((formalparameters)) .
;        (list of sequential elements, maybe empty))
; element: (list of alternatives) or terminal-pattern or predicate
; terminal-pattern: (terminal) or (terminal "value")
; alternative: ((actualparameters) rule)
; predicate: ((actualparameters) predicate-name)
; parameters: list of in and out: ((in-params) (out-params))

; when loaded, a list of attributes is appended to each
; (cdadr rule), i.e. next to the formal parameters.
; attributes: (emptyness, firstset, followset)
; compilationunit: (car grammar)

; the output is a syntax tree with nodes of the form:
; (rule-name formal-parameters . (list of sub-trees and terminals-as-matched))
; a sub-tree has the actual parameters prepended:
; (actual-parameters . syntax-tree)


(load "pretty-print.so")
(define debug-grammar #t)   ;do a check for validity of rules

(define write-error
  (lambda (class source text)
    (write
      (list
        class
        (list
          (cons 'source source)
          '(stage parse))
        text)
      (current-error-port))
    (newline (current-error-port))
    #f))

;;; a few functions to extract a predicate from the attributes (cdadr rule):
; is the emptiness-flag set for the rule?
(define rule-maybe-empty?
  (lambda (rule)
    (cadadr rule)))

;;; the first-set as calculated for the rule
(define rule-first-set
  (lambda (rule)
    (cadr (cdadr rule))))

;;; the follow-set as calculated for the rule
(define rule-follow-set
  (lambda (rule)
    (caddr (cdadr rule))))


;;; a few functions to determine the nature of a rule-item:
; is the rule-item a set of alternative nonterminals ?
(define item-alternatives?
  (lambda (item)
    (and
      (list? item)   ;list of alternatives
      (list? (car item))   ;alternative a list of parameters + nonterm
      (list? (caar item))   ;parameters a list
      (or
        (null? (cdr item))   ;either only one alternative
        (list? (cadr item))))))   ;or another alternative

; is the rule-item a predicate ?
(define item-predicate?
  (lambda (item)
    (and
      (list? item)   ;list of parameters + predicate-name
      (list? (car item))   ;parameters a list
      (not (null? (cdr item)))   ;predicate-name obligatory
      (symbol? (cadr item)))))   ;predicate-name a symbol

; is the rule-item a terminal ?
(define item-terminal?
  (lambda (item)
    (and
      (list? item)   ;pattern
      (symbol? (car item)))))   ;terminal-name a symbol

;;; a few functions to cope with terminal symbols:
; does terminal b match terminal trunk pattern a?
(define terminal-head-of?
  (lambda (a b)
    (or
      (null? a)   ;empty set is head of anything: #t
      (and
        (not (null? b))    ;but nothing else is head of an empty set: #f
        (equal? (car a) (car b))
        (terminal-head-of? (cdr a) (cdr b))))))

;;; is terminal a less than terminal b?
(define terminal-less-than?
  (lambda (a b)
    (and
      (not (null? b))   ;nothing is less than the empty set: #f
      (or
        (null? a)   ;empty set is less than anything else: #t
        (let ((ha (car a)) (hb (car b)))
          (if (equal? ha hb)   ;if head is equal, compare the rest
            (terminal-less-than? (cdr a) (cdr b))
            (string<?
              (if (symbol? ha) (symbol->string ha) ha)
              (if (symbol? hb) (symbol->string hb) hb))))))))


;;; unify a set (b) into another one (a).
; when done, call lambda "action" with parameters "result" and "changed"
(define prepare-unify-set
  (lambda (action result changed set-a set-b)
    (if (null? set-b)   ;nothing more to be merged: skip the rest of set-a
      (if (null? set-a)
        (action result changed)
        (prepare-unify-set action (append result set-a) changed '() '()))
      (if (null? set-a)   ;mergeables left, but set-a done: append the rest
        (prepare-unify-set action (append result set-b) #t '() '())
        (if (terminal-head-of? (car set-a) (car set-b))
          (prepare-unify-set action result changed set-a (cdr set-b))
          (if (terminal-head-of? (car set-b) (car set-a))
            (prepare-unify-set action result changed (cdr set-a) set-b)
            (if (terminal-less-than? (car set-a) (car set-b))
              (prepare-unify-set   ;just take the next from a and procede
                action
                (append result (list (car set-a)))
                changed
                (cdr set-a)
                set-b)
              (prepare-unify-set   ;the next from b is unique: this is a change!
                action
                (append result (list (car set-b)))
                #t
                set-a
                (cdr set-b)))))))))
(define prepare-add-set
  (lambda (set-at-car merge-this)
    (prepare-unify-set
      (lambda (result changed)
        (if changed (set-car! set-at-car result))
        changed)
      '()
      #f
      (car set-at-car)
      merge-this)))

;;; determine the follow-sets for all rules
(define prepare-follow-follow
  (lambda (grammar rule item maybe-empty items)
    (if (null? item)   ;first called, or one alternatives done
      (if maybe-empty
        (if (null? items)
          (rule-follow-set rule)   ;last item: follow(i) >= follow(r)
          (if (item-alternatives? (car items))
            (prepare-follow-follow grammar rule (car items) #f (cdr items))
            (if (item-predicate? (car items))
              (prepare-follow-follow grammar rule '() #t (cdr items))
              (list (car items)))))   ; follow(i) >= {term(i+1)}
        '())
      (let ((nonterm (assq (cadar item) grammar)))
        (prepare-unify-set   ; follow(i) >= unify(all first(i+1))
          (lambda (result changed) result)
          '()
          #f
          (rule-first-set nonterm)
          (prepare-follow-follow
            grammar
            rule
            (cdr item)
            (or maybe-empty (rule-maybe-empty? nonterm))
            items))))))
(define prepare-follow-item
  (lambda (grammar rule rules item items changed)
    (if
      (or   ;if alternatives done, or if not alternatives
        (null? item)
        (not (item-alternatives? item)))
      (if (null? items)   ;rule done: next rule
        (prepare-follow-rule grammar rules changed)
        (prepare-follow-item   ;...else proceed with next item in rule
          grammar
          rule
          rules
          (car items)
          (cdr items)
          changed))
      (prepare-follow-item
        grammar
        rule
        rules
        (cdr item)   ;next alternative in item
        items
        (or
          (prepare-add-set   ;add new follow to follow-set of nonterm mentioned
            (cdddr (cadr (assq (cadar item) grammar)))
            (prepare-follow-follow grammar rule '() #t items))
          changed)))))
(define prepare-follow-rule
  (lambda (grammar rules changed)
    (if (null? rules)
      (if changed
        (prepare-follow-sets grammar)   ;just do it once more
        grammar)
      (if (null? (cddar rules))   ;empty rule: nothing to do
        (prepare-follow-rule grammar (cdr rules) changed)
        (prepare-follow-item
          grammar
          (car rules)   ;current rule
          (cdr rules)   ;remaining rules
          (caddar rules)   ;first item
          (cdddar rules)   ;following items
          changed)))))
(define prepare-follow-sets
  (lambda (grammar)
    (prepare-follow-rule grammar grammar #f)))

;;; determine the first-sets for all rules
(define prepare-first-alternative
  (lambda (grammar rule rules item maybe-empty items changed)
    (if (null? item)   ;if alternatives exhausted and...
      (if
        (and maybe-empty (not (null? items)))
        (prepare-first-item grammar rule rules items changed)
        (prepare-first-rule grammar rules changed))
      (let ((nonterm (assq (cadar item) grammar)))
        (prepare-first-alternative
          grammar
          rule
          rules
          (cdr item)
          (or maybe-empty (rule-maybe-empty? nonterm))
          items
          (or
            (prepare-add-set (cddadr rule) (rule-first-set nonterm))
            changed))))))
(define prepare-first-item
  (lambda (grammar rule rules items changed)
    (if (null? items)   ;empty rule: nothing to do
      (prepare-first-rule grammar rules changed)
      (if (item-alternatives? (car items))
        (prepare-first-alternative
          grammar
          rule
          rules
          (car items)
          #f
          (cdr items)
          changed)
        (if (item-predicate? (car items))
          (prepare-first-item grammar rule rules (cdr items) changed)
          (prepare-first-rule
            grammar
            rules
            (or
              (prepare-add-set (cddadr rule) (list (car items)))
              changed)))))))
(define prepare-first-rule
  (lambda (grammar rules changed)
    (if (null? rules)
      (if changed
        (prepare-first-sets grammar)   ;just do it once more
        grammar)   ;all rules traversed without change: done
      (prepare-first-item
        grammar
        (car rules)
        (cdr rules)
        (cddar rules)
        changed))))
(define prepare-first-sets
  (lambda (grammar)
    (prepare-first-rule grammar grammar #f)))

;;; determine potential emptiness of all rules
(define prepare-empty-item
  (lambda (grammar item)
    (and
      (not (null? item))   ;none of the alternatives maybe empty: cannot
      (or
        (rule-maybe-empty? (assq (cadar item) grammar))
        (prepare-empty-item grammar (cdr item))))))
(define prepare-empty-items
  (lambda (grammar items)
    (or
      (null? items)  ;no more items left: maybe obviously empty
      (and
        (or
          (item-predicate? (car items))   ;predicate is always empty
          (and
            (item-alternatives? (car items))
            (prepare-empty-item grammar (car items))))
        (prepare-empty-items grammar (cdr items))))))
(define prepare-empty-rule
  (lambda (grammar rules changed)
    (if (null? rules)
      (if changed
        (prepare-emptinesses grammar)
        grammar)
      (prepare-empty-rule
        grammar
        (cdr rules)
        (if
          (and
            (not (rule-maybe-empty? (car rules)))   ;flag set yet: done
            (prepare-empty-items grammar (cddar rules)))
          (begin
            (set-car! (cdadar rules) #t)   ;set the "empty" flag
            #t)
          changed)))))
(define prepare-emptinesses
  (lambda (grammar)
    (prepare-empty-rule grammar grammar #f)))

;;; add an empty attributes field to the end of each (cadr rule):
; '(emptiness first-set follow-set)
(define prepare-addfields
  (lambda (result grammar)
    (if (null? grammar) result
      (prepare-addfields
        (append result
          (list (append
            (list (caar grammar))
            (cons
              (append (cadar grammar) (list #f '() '()))
              (cddar grammar)))))
        (cdr grammar)))))


;;; check that all referenced nonterminals do exist
(define grammar-complete-alternatives
  (lambda (grammar item)
    (or
      (null? item)
      (if
        (assq (cadar item) grammar)
        (grammar-complete-alternatives grammar (cdr item))
        (begin
          (write-error
            'config
            '(0 0)
            "undefined nonterminal in grammar")
          (write-error
            'debug
            '(0 0)
            (cadar item))
          #f)))))
(define grammar-complete-items
  (lambda (grammar rule items)
    (or
      (null? items)
      (if
        (null? (car items))
        (begin
          (write-error
            'config
            '(0 0)
            "alternatives must not be empty")
          (write-error
            'debug
            '(0 0)
            (car rule))
          #f)
        (and
          (or
            (not (item-alternatives? (car items)))
            (grammar-complete-alternatives grammar (car items)))
          (grammar-complete-items grammar rule (cdr items)))))))
(define grammar-complete
  (lambda (grammar rules)
    (or
      (null? rules)
      (and
        (grammar-complete-items grammar (car rules) (cddar rules))
        (grammar-complete grammar (cdr rules))))))


;;; check that the first- and follow-sets are usable
(define debug-sets-alternative
  (lambda (grammar item unity count)
    (if (null? item)
      (equal? (length unity) count)
      (let ((f (rule-first-set (assq (cadar item) grammar))))
        (debug-sets-alternative
          grammar
          (cdr item)
          (prepare-unify-set
            (lambda (result changed) result)
            '()
            #f
            unity
            f)
          (+ count (length f)))))))
(define debug-sets-alternatives
  (lambda (grammar rule items)
    (or
      (null? items)
      (and
        (or
          (not (item-alternatives? (car items)))
          (debug-sets-alternative grammar (car items) '() 0)
          (begin
            (write-error
              'config
              '(0 0)
              "first/follow sets not unique for alternatives")
            (write-error
              'debug
              '(0 0)
              rule)
            #f))
        (debug-sets-alternatives grammar rule (cdr items))))))
(define debug-sets-first-follow
  (lambda (rule)
    (or
      (not (rule-maybe-empty? rule))
      (equal?
        (+ (length (rule-first-set rule)) (length (rule-follow-set rule)))
        (length
          (prepare-unify-set
            (lambda (result changed) result)
            '()
            #f
            (rule-first-set rule)
            (rule-follow-set rule))))
      (begin
        (write-error
          'config
          '(0 0)
          "first/follow sets not unique for possibly empty rule")
        (write-error
          'debug
          '(0 0)
          rule)
        #f))))
(define debug-sets
  (lambda (grammar rules)
    (or (null? rules)
      (and
        (debug-sets-alternatives grammar (car rules) (cddar rules))
        (debug-sets-first-follow (car rules))
        (debug-sets grammar (cdr rules))))))


;;; check that no circulair replacement can be done by selecting the first
; alternative allways. the first alternative shall be the default when
; encountering an unexpected token, so its evaluation must terminate.
(define first-noncirculair-items
  (lambda (grammar done items stack)
    (if (null? items) done   ;return nonterminals done, collected
      (if (item-alternatives? (car items))
        (let
          ((d
            (first-noncirculair-rule
              grammar
              done
              (assq (cadaar items) grammar)
              stack)))
          (if d
            (first-noncirculair-items grammar d (cdr items) stack)
            #f))
        (first-noncirculair-items grammar done (cdr items) stack)))))
(define first-noncirculair-rule
  (lambda (grammar done rule stack)
    (if (assq (car rule) stack)
      (begin
        (write-error
          'config
          '(0 0)
          "grammar must be non-circulair in first alternatives")
        (write-error
          'debug
          '(0 0)
          (cons (list (car rule)) stack))
        #f)
      (first-noncirculair-items
        grammar
        (cons (list (car rule)) done)
        (cddr rule)
        (cons (list (car rule)) stack)))))
(define first-noncirculair
  (lambda (grammar rules done)
    (and done   ;if false -> found a circle
      (or
        (null? rules)   ;if empty -> finished
        (first-noncirculair
          grammar
          (cdr rules)   ;check next rule, but first...
          (if (assq (caar rules) done)   ;checked yet -> skip rule
            done
            (first-noncirculair-rule grammar done (car rules) '())))))))


;;; from the global list of following items compute the
; recovery set of tokens to skip to
(define recovery-set-alternative
  (lambda (grammar items alternatives)
    (if (null? alternatives)
      (recovery-set-items grammar items)
      (prepare-unify-set
        (lambda (result changed) result)
        '()
        #f
        (rule-first-set (assq (cadar alternatives) grammar))
        (recovery-set-alternative grammar items (cdr alternatives))))))
(define recovery-set-items
  (lambda (grammar items)
    (if (null? items)
      '()
      (if (item-alternatives? (car items))
        (recovery-set-alternative grammar (cdr items) (car items))
        (if (item-predicate? (car items))
          (recovery-set-items grammar (cdr items))
          (prepare-unify-set
            (lambda (result changed) result)
            '()
            #f
            (list (car items))
            (recovery-set-items grammar (cdr items))))))))


(define parse-token-in-set?
  (lambda (token set)
    (and
      (not (null? set))
      (or
        (terminal-head-of? (car set) token)
        (parse-token-in-set? token (cdr set))))))
(define parse-skip-to-match
  (lambda (grammar token recoveryset)
    (if (list? token)
      (if (parse-token-in-set? token recoveryset)
        (begin
          (write-error
            'repair
            (caddr token)
            "repair mode restart point")
          (write-error
            'debug
            (caddr token)
            token)
          token)
        (parse-skip-to-match grammar (read) recoveryset))
      (begin
        (write-error
          'fatal
          '(0 0)
          "unexpected end of file at skip-to-match")
        token))))
(define parse-alternatives
  (lambda (grammar rule item alternatives token error recovery)
    (if (null? alternatives)
      (begin
        (write-error
          'error
          (caddr token)
          "unexpected token at alternatives")
        (write-error
          'debug
          (caddr token)
          token)
        (write-error
          'debug
          (caddr token)
          (list "item:" item))
;; on token mismatch, collect the global recovery set (unification of all
;; first-sets of all successors of the current item), skip tokens up to
;; the first within the global recovery set, then skip items until that
;; token matches (the latter step comes gratuit, but: "until" may be immediate,
;; so recheck the current alternatives for the fresh token):
        (let
          ((t
            (parse-skip-to-match
              grammar
              token
              (recovery-set-items grammar recovery))))
          (if (equal? t token)   ;compare, including position
            (parse-rule   ;when nothing skipped: default is first alternative
              grammar
              (assq (cadar item) grammar)   ;rule
              (caar item)   ;actual parameters
              token
              #t
              recovery)
            (parse-alternatives   ;when some token skipped:
              grammar
              rule
              item
              item   ;try all alternatives again, with...
              t   ;...fresh token from skip
              #t
              recovery))))
      (if (list? token)
        (let ((r (assq (cadar alternatives) grammar)))
          (if
            (or
              (parse-token-in-set? token (rule-first-set r))
              (and
                (rule-maybe-empty? r)
                (parse-token-in-set? token (rule-follow-set r))))
            (parse-rule grammar r (caar alternatives) token error recovery)
            (parse-alternatives
              grammar
              rule
              item   ;the complete alternatives, used in case of error
              (cdr alternatives)   ;the remaining alternatives
              token
              error
              recovery)))
        (values '() token error)))))
(define parse-terminal
  (lambda (grammar rule terminal token error)
    (if (list? token)
      (if (terminal-head-of? terminal token)
        (values token (read) error)
        (begin
          (write-error
            'error
            (caddr token)
            "unexpected token at terminal")
          (write-error
            'debug
            (caddr token)
            token)
          (write-error
            'debug
            (caddr token)
            (list "terminal:" terminal))
          (values '() token (caddr token))))
      (begin
        (write-error
          'fatal
          '(0 0)
          "unexpected end of file at terminal")
        (values '() token '(0 . 0))))))
(define parse-items
  (lambda (grammar rule items sequence token error recovery)
    (if (null? items)
      (values sequence token error)
      (call-with-values
        (lambda ()
          (if (item-alternatives? (car items))
            (parse-alternatives
              grammar
              rule   ;the rule we are in
              (car items)   ;the expected alternatives
              (car items)   ;the remaining alternatives
              token   ;the current token
              error
              (append   ;all items relevant for recovery set, simply collected
                (cdr items)
                recovery))
            (if (item-predicate? (car items))
              (values (car items) token error)   ;simply emit the predicate
              (parse-terminal
                grammar
                rule   ;the rule we are in
                (car items)   ;the expected terminal
                token   ;the current token
                error))))
        (lambda (tree token error)
          (parse-items
            grammar
            rule
            (cdr items)
            (append sequence (list tree))
            token
            error
            recovery))))))
(define parse-tree
  (lambda (grammar rule token error recovery)
    (call-with-values
      (lambda ()
        (parse-items grammar rule (cddr rule) '() token error recovery))
      (lambda (sequence token error)
        (values
          (cons
            (car rule)   ;the rule's name
            (cons
              (caadr rule)   ;the formal parameters
              sequence))
          token
          error)))))
(define parse-rule
  (lambda (grammar rule actual-params token error recovery)
    (call-with-values
      (lambda () (parse-tree grammar rule token error recovery))
      (lambda (tree token error)
        (values
          (cons actual-params tree)
          token
          error)))))
(define parse-file
  (lambda (grammar)
    (call-with-values
      (lambda ()
        (parse-tree
          grammar   ;the grammar goes everywhere
          (car grammar)   ;the first rule is the compilation unit
          (read)   ;start at the first token
          #f   ;no error yet
          '()))   ;the list of following items, relevant for recovery-set
      (lambda (tree token error)
        (if (list? token)
          (begin
            (write-error
              'fatal
              (caddr token)
              "superfluous token")
            (write-error
              'debug
              (caddr token)
              token))
          (if error
            #f
            tree))))))
          

(define extended-affix-grammar (make-parameter #f))
(load (cadr (argv)))
(if
  (or
    (not debug-grammar)
    (grammar-complete (extended-affix-grammar) (extended-affix-grammar)))
  (let ((grammar
      (prepare-follow-sets
        (prepare-first-sets
          (prepare-emptinesses
            (prepare-addfields '() (extended-affix-grammar)))))))
    (if
      (or
        (not debug-grammar)
        (and
          (debug-sets grammar grammar)
          (first-noncirculair grammar grammar '())))
      (pretty-print (parse-file grammar)))))
(newline)
